@@ -2,7 +2,7 @@ fn main() {
     let src_dir = std::path::Path::new("src");
 
     let mut c_config = cc::Build::new();
-    c_config.std("c11").include(src_dir).flag_if_supported("-Wno-unused-value");
+    c_config.std("c11").include(src_dir);
 
     #[cfg(target_env = "msvc")]
     c_config.flag("-utf-8");
@@ -12,8 +12,10 @@ fn main() {
     println!("cargo:rerun-if-changed={}", parser_path.to_str().unwrap());
 
     let scanner_path = src_dir.join("scanner.c");
-    c_config.file(&scanner_path);
-    println!("cargo:rerun-if-changed={}", scanner_path.to_str().unwrap());
+    if scanner_path.exists() {
+        c_config.file(&scanner_path);
+        println!("cargo:rerun-if-changed={}", scanner_path.to_str().unwrap());
+    }
 
     c_config.compile("tree-sitter-julia");
 }
